using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Printing;
using System.Drawing.Text;
using System.Text;
using System.Windows.Forms;

namespace IntSight.Controls
{
    using IntSight.Controls.CodeModel;

    /// <summary>Prints the document managed by a code editor.</summary>
    public class EditorDocument : PrintDocument
    {
        private CodeEditor editor;
        private IEnumerator<Lexeme> tokenizer;
        private StringFormat stringFormat;
        private Brush textBrush;
        private Font normalFont, italicFont, boldFont, activeFont;
        private int pageNumber, lineNumber;
        private float lineHeight, xPos, yPos;
        private bool lineNumbers;

        public EditorDocument()
        {
            this.stringFormat = new StringFormat(StringFormat.GenericTypographic);
            this.stringFormat.FormatFlags |= StringFormatFlags.MeasureTrailingSpaces;
            this.ResetFont();
            this.lineNumbers = true;
        }

        [Browsable(true)]
        [Description("The code editor providing the content to be printed.")]
        public CodeEditor Editor
        {
            get { return editor; }
            set { editor = value; }
        }

        [Browsable(true)]
        [Description("Show line numbers along the left margin.")]
        [DefaultValue(true)]
        public bool LineNumbers
        {
            get { return lineNumbers; }
            set { lineNumbers = value; }
        }

        [Browsable(true)]
        [Description("Base font used to print text.")]
        public Font Font
        {
            get { return normalFont; }
            set { normalFont = value; }
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        private bool ShouldSerializeFont()
        {
            return String.Compare(this.Font.FontFamily.Name, "Courier New",
                StringComparison.InvariantCultureIgnoreCase) != 0 ||
                this.Font.SizeInPoints != 9.0F ||
                this.Font.Style != 0;
        }

        [EditorBrowsable(EditorBrowsableState.Never)]
        private void ResetFont()
        {
            if (this.Font != null)
                this.Font.Dispose();
            this.Font = new Font("Courier New", 9.0F);
        }

        protected override void OnBeginPrint(PrintEventArgs e)
        {
            if (editor != null)
            {
                italicFont = new Font(normalFont, FontStyle.Italic);
                boldFont = new Font(normalFont, FontStyle.Bold);
                activeFont = normalFont;
                textBrush = Brushes.Black;
                pageNumber = 0;
                lineNumber = 0;
                tokenizer = editor.Tokens().GetEnumerator();
            }
            base.OnBeginPrint(e);
        }

        protected override void OnEndPrint(PrintEventArgs e)
        {
            base.OnEndPrint(e);
            if (editor != null)
            {
                boldFont.Dispose();
                italicFont.Dispose();
                tokenizer.Dispose();
                tokenizer = null;
                activeFont = italicFont = boldFont = null;
                textBrush = null;
            }
        }

        protected override void OnPrintPage(PrintPageEventArgs e)
        {
            if (editor != null)
            {
                e.Graphics.TextRenderingHint = TextRenderingHint.AntiAlias;
                float leftMargin = e.MarginBounds.Left;
                float topMargin = e.MarginBounds.Top;

                // Calculate the number of lines per page.
                lineHeight = normalFont.GetHeight(e.Graphics);
                float linesPerPage = e.MarginBounds.Height / lineHeight;
                // Print a header.
                stringFormat.Alignment = StringAlignment.Near;
                stringFormat.Trimming = StringTrimming.EllipsisPath;
                e.Graphics.DrawString(
                    DocumentName, normalFont, Brushes.Black,
                    new RectangleF(
                        leftMargin, topMargin, e.MarginBounds.Width - 20, lineHeight),
                    stringFormat);
                stringFormat.Trimming = StringTrimming.None;
                stringFormat.Alignment = StringAlignment.Far;
                e.Graphics.DrawString((++pageNumber).ToString(),
                    normalFont, Brushes.Black,
                    new RectangleF(
                        leftMargin, topMargin, e.MarginBounds.Width, lineHeight),
                    stringFormat);
                e.Graphics.DrawLine(Pens.Black,
                    leftMargin, topMargin + lineHeight + 2,
                    e.MarginBounds.Right, topMargin + lineHeight + 2);
                linesPerPage -= 2;
                // Print each line of the file.
                stringFormat.Trimming = StringTrimming.EllipsisCharacter;
                stringFormat.Alignment = StringAlignment.Near;
                int lineNo = 0;
                xPos = leftMargin;
                yPos = topMargin + 2 * lineHeight;
                bool atFirstCharacter = true;
                while (true)
                {
                    if (!tokenizer.MoveNext())
                    {
                        base.OnPrintPage(e);
                        e.HasMorePages = false;
                        return;
                    }
                    Lexeme lexeme = tokenizer.Current;
                    if (atFirstCharacter && lineNumbers)
                    {
                        PrintLineNumber(e);
                        atFirstCharacter = false;
                    }
                    switch (lexeme.Kind)
                    {
                        case Lexeme.Token.Text:
                        case Lexeme.Token.Error:
                            PrintText(e, lexeme.Text, normalFont, Brushes.Black);
                            break;
                        case Lexeme.Token.Keyword:
                            PrintText(e, lexeme.Text, boldFont, Brushes.Navy);
                            break;
                        case Lexeme.Token.String:
                            PrintText(e, lexeme.Text, normalFont, Brushes.Maroon);
                            break;
                        case Lexeme.Token.PartialComment:
                        case Lexeme.Token.Comment:
                            PrintText(e, lexeme.Text, italicFont, Brushes.Green);
                            break;
                        case Lexeme.Token.NewLine:
                            lineNo++;
                            lineNumber++;
                            atFirstCharacter = true;
                            xPos = leftMargin;
                            yPos += lineHeight;
                            if (lineNo >= linesPerPage)
                            {
                                base.OnPrintPage(e);
                                e.HasMorePages = true;
                                return;
                            }
                            break;
                    }
                }
            }
            base.OnPrintPage(e);
        }

        private void PrintText(PrintPageEventArgs e, string text, Font font, Brush brush)
        {
            if (xPos < e.MarginBounds.Right)
            {
                RectangleF rect = new RectangleF(
                    xPos, yPos, e.MarginBounds.Right - xPos, lineHeight);
                e.Graphics.DrawString(text, font, brush, rect, stringFormat);
                xPos += e.Graphics.MeasureString(text, font,
                    new PointF(0, 0), stringFormat).Width;
            }
        }

        private void PrintLineNumber(PrintPageEventArgs e)
        {
            string lineText = (lineNumber + 1).ToString() + " ";
            stringFormat.Alignment = StringAlignment.Far;
            stringFormat.Trimming = StringTrimming.None;
            e.Graphics.DrawString(lineText, normalFont, Brushes.DimGray,
                new RectangleF(0, yPos, e.MarginBounds.Left, lineHeight), stringFormat);
            stringFormat.Alignment = StringAlignment.Near;
            stringFormat.Trimming = StringTrimming.EllipsisCharacter;
        }
    }
}
