using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using IntSight.Controls;

namespace hammer
{
    public partial class FindDialog : Form
    {
        private static FindDialog instance = null;
        private IntSight.Controls.CodeEditor editor;
        private FindCompleteEventHandler saveFindComplete;

        public static void Show(
            IntSight.Controls.CodeEditor editor,
            FindCompleteEventHandler findComplete)
        {
            if (instance == null)
            {
                ReplaceDialog.Shutdown();
                instance = new FindDialog();
                instance.editor = editor;
                string text = editor.SelectedText;
                if (!string.IsNullOrEmpty(text))
                    instance.textCombo.Text = text;
                instance.saveFindComplete = findComplete;
                if (findComplete != null)
                    instance.FindComplete += findComplete;
                instance.Show(editor.FindForm());
            }
        }

        public static void Beep()
        {
            using (System.Media.SoundPlayer sp =
                new System.Media.SoundPlayer(hammer.Properties.Resources.notFound))
                sp.Play();
        }

        public static void SetOpacity()
        {
            if (instance != null)
                instance.Opacity = MainForm.GlobalOpacity;
        }

        public static void Shutdown()
        {
            if (instance != null)
                instance.Close();
        }

        public event FindCompleteEventHandler FindComplete;

        private FindDialog()
        {
            InitializeComponent();
            Opacity = (MainForm.GlobalOpacity - 1) / 100.0;
            RestoreValues();
            FormClosed += FindForm_FormClosed;
        }

        private void SaveValues()
        {
            MainForm.WriteRegistry("Find text", textCombo.Text);
            MainForm.WriteRegistry("Match case", matchCase.Checked);
            MainForm.WriteRegistry("Whole words", wholeWords.Checked);
            MainForm.WriteRegistry("Find left position", Left);
            MainForm.WriteRegistry("Find top position", Top);
            string[] findStrings = new string[textCombo.Items.Count];
            textCombo.Items.CopyTo(findStrings, 0);
            MainForm.WriteRegistry("Find items", findStrings);
        }

        private void RestoreValues()
        {
            textCombo.Text = MainForm.Read("Find text", String.Empty);
            matchCase.Checked = MainForm.Read("Match case", false);
            wholeWords.Checked = MainForm.Read("Whole words", false);
            int leftPos = MainForm.Read("Find left position", -1);
            int topPos = MainForm.Read("Find top position", -1);
            if (leftPos >= 0 && topPos >= 0)
            {
                this.StartPosition = FormStartPosition.Manual;
                this.Location = new Point(leftPos, topPos);
            }
            textCombo.Items.AddRange((string[])MainForm.ReadRegistry(
                "Find items", new string[0]));
        }

        private void FindForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            SaveValues();
            if (saveFindComplete != null)
                FindComplete -= saveFindComplete;
            instance = null;
        }

        private void bnFind_Click(object sender, EventArgs e)
        {
            string findString = textCombo.Text;
            textCombo.SelectAll();
            int pos = textCombo.Items.IndexOf(findString);
            if (pos != -1)
                textCombo.Items.RemoveAt(pos);
            if (textCombo.Items.Count == 8)
                textCombo.Items.RemoveAt(7);
            textCombo.Items.Insert(0, findString);
            bool result = editor.FindText(
                findString, editor.Current, wholeWords.Checked, matchCase.Checked);
            if (!result && editor.Current > CodeEditor.Position.Zero)
                result = editor.FindText(
                    findString, CodeEditor.Position.Zero, wholeWords.Checked, matchCase.Checked);
            OnFindComplete(result ?
                FindCompleteEventArgs.Found : FindCompleteEventArgs.NotFound);
            if (!result)
                FindDialog.Beep();
        }

        protected virtual void OnFindComplete(FindCompleteEventArgs e)
        {
            FindCompleteEventHandler eh = FindComplete;
            if (eh != null)
                eh(this, e);
        }

        private void bnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void FindForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((e.Modifiers & (Keys.Control | Keys.Shift | Keys.Alt)) == Keys.Control &&
                e.KeyCode == Keys.R)
            {
                ReplaceDialog.Show(editor, FindComplete);
                e.Handled = true;
            }
        }
    }

    public delegate void FindCompleteEventHandler(object sender, FindCompleteEventArgs e);

    public enum FindResult { NotFound, Found, Replaced, ReplacedAll }

    public sealed class FindCompleteEventArgs : EventArgs
    {
        public static readonly FindCompleteEventArgs Found =
            new FindCompleteEventArgs(FindResult.Found);
        public static readonly FindCompleteEventArgs NotFound =
            new FindCompleteEventArgs(FindResult.NotFound);
        public static readonly FindCompleteEventArgs Replaced =
            new FindCompleteEventArgs(FindResult.Replaced);

        public readonly FindResult Reason;
        public readonly int Count;

        public FindCompleteEventArgs(FindResult reason)
        {
            this.Reason = reason;
            if (reason != FindResult.NotFound)
                this.Count = 1;
        }

        public FindCompleteEventArgs(int count)
        {
            this.Reason = FindResult.ReplacedAll;
            this.Count = count;
        }
    }
}