using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32;

namespace hammer
{
    public partial class MainForm : Form
    {
        public const string BaseKey = @"HKEY_CURRENT_USER\Software\Intuitive Sight\Freya Pad\1.0\Settings";
        public const string KeyMru = "MRU Files";
        public const string KeyVisualStyle = "Visual style";
        public const string KeyOpacity = "Opacity";
        public const string KeyEditorMargin = "Editor margin";
        public const string KeyTabSize = "Tabs";

        private const string DefaultTitle = "Freya Pad";
        private MruList mruList;
        private string fileName;
        private static int globalOpacity = 100;
        private static MainForm instance;

        public MainForm()
        {
            instance = this;
            InitializeComponent();
            switch (MainForm.Read(KeyVisualStyle, 0))
            {
                case 1: MainForm.VStyle = VisualStyle.Tan; break;
                case 2: MainForm.VStyle = VisualStyle.Silver; break;
                default: MainForm.VStyle = VisualStyle.Blue; break;
            }
            codeEditor.CodeScanner = new FreyaScanner();
            MainForm.GlobalOpacity = MainForm.Read(KeyOpacity, 100);
            codeEditor.KeywordColor = MainForm.Read("Keywords color", codeEditor.KeywordColor);
            codeEditor.CommentColor = MainForm.Read("Comments color", codeEditor.CommentColor);
            codeEditor.StringColor = MainForm.Read("Strings color", codeEditor.StringColor);
            MainForm.TabSize = MainForm.Read(KeyTabSize, 4);
            MainForm.LeftMargin = MainForm.Read(KeyEditorMargin, 12);
            string fontName = MainForm.Read("Editor font", codeEditor.Font.Name);
            string fontSize = MainForm.Read("Editor font size", codeEditor.Font.Size.ToString());
            float fSize;
            if (Single.TryParse(fontSize, out fSize))
                MainForm.EditorFont = new Font(fontName, fSize);
            FileName = String.Empty;
            mruList = new MruList(8, BaseKey, KeyMru, miReopen);
            mruList.FileOpen += mruList_FileOpen;
            Application.Idle += Application_Idle;
            codeEditor.SnippetManager = new SnippetManager(Application.StartupPath);
        }

        public static object ReadRegistry(string valueName, object defaultValue)
        {
            return Registry.GetValue(BaseKey, valueName, defaultValue) ?? defaultValue;
        }

        public static string Read(string valueName, string defaultValue)
        {
            return Convert.ToString(ReadRegistry(valueName, defaultValue));
        }

        public static bool Read(string valueName, bool defaultValue)
        {
            return Convert.ToBoolean(ReadRegistry(valueName, defaultValue));
        }

        public static int Read(string valueName, int defaultValue)
        {
            return Convert.ToInt32(ReadRegistry(valueName, defaultValue));
        }

        public static void WriteRegistry(string valueName, object value)
        {
            Registry.SetValue(BaseKey, valueName, value);
        }

        private static Color Read(string valueName, Color defaultColor)
        {
            return Color.FromArgb(Read(valueName, defaultColor.ToArgb()));
        }

        #region Static properties.

        public enum VisualStyle { Blue, Tan, Silver }

        private static VisualStyle visualStyle;

        public static VisualStyle VStyle
        {
            get { return visualStyle; }
            set
            {
                ToolStripManager.RenderMode = ToolStripManagerRenderMode.Professional;
                ToolStripManager.VisualStylesEnabled = true;
                switch (value)
                {
                    case VisualStyle.Blue:
                        ToolStripManager.Renderer = new ToolStripProfessionalRenderer();
                        instance.codeEditor.MarginColor = Color.LightSteelBlue;
                        break;
                    case VisualStyle.Silver:
                        ToolStripManager.Renderer = new ToolStripProfessionalRenderer(
                            new SilverColorTable());
                        instance.codeEditor.MarginColor = Color.Gainsboro;
                        break;
                    case VisualStyle.Tan:
                        ToolStripManager.Renderer = new ToolStripProfessionalRenderer(
                            new TanColorTable());
                        instance.codeEditor.MarginColor = Color.FromArgb(0xe5, 0xe5, 0xd7);
                        break;
                }
                visualStyle = value;
            }
        }

        public static int GlobalOpacity
        {
            get { return globalOpacity; }
            set
            {
                globalOpacity = value < 50 ? 50 : value > 100 ? 100 : value;
                FindDialog.SetOpacity();
                ReplaceDialog.SetOpacity();
            }
        }

        public static Color KeywordColor
        {
            get { return instance.codeEditor.KeywordColor; }
            set { instance.codeEditor.KeywordColor = value; }
        }

        public static Color CommentColor
        {
            get { return instance.codeEditor.CommentColor; }
            set { instance.codeEditor.CommentColor = value; }
        }

        public static Color StringColor
        {
            get { return instance.codeEditor.StringColor; }
            set { instance.codeEditor.StringColor = value; }
        }

        public static int TabSize
        {
            get { return instance.codeEditor.TabLength; }
            set { instance.codeEditor.TabLength = value; }
        }

        public static int LeftMargin
        {
            get { return instance.codeEditor.LeftMargin; }
            set { instance.codeEditor.LeftMargin = value; }
        }

        public static Font EditorFont
        {
            get { return instance.codeEditor.Font; }
            set
            {
                Font oldFont = instance.codeEditor.Font;
                if (oldFont != value)
                {
                    instance.codeEditor.Font = value;
                    oldFont.Dispose();
                }
            }
        }

        #endregion Static properties.

        #region Form lifetime.

        private void Application_Idle(object sender, EventArgs e)
        {
            lineLabel.Text = String.Format("Line {0:#,0}", codeEditor.CurrentLine);
            columnLabel.Text = String.Format("Column {0:#,0}", codeEditor.CurrentColumn);
            miFolder.Enabled = !String.IsNullOrEmpty(fileName);
            miSave.Enabled = bnSave.Enabled = codeEditor.Modified;
            miCopy.Enabled = miCut.Enabled = miDelete.Enabled =
                miComment.Enabled = miUncomment.Enabled =
                bnComment.Enabled = bnUncomment.Enabled =
                miUpper.Enabled = miLower.Enabled =
                bnCopy.Enabled = bnCut.Enabled = codeEditor.HasSelection;
            miPaste.Enabled = bnPaste.Enabled = codeEditor.CanPaste;
            miUndo.Enabled = bnUndo.Enabled = codeEditor.CanUndo;
            miRedo.Enabled = bnRedo.Enabled = codeEditor.CanRedo;
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            e.Cancel = !CanClose();
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            Application.Idle -= Application_Idle;
        }

        private bool CanClose()
        {
            if (!codeEditor.Modified)
                return true;
            string message = String.IsNullOrEmpty(FileName) ?
                "This file" : Path.GetFileName(FileName);
            DialogResult rslt = MessageBox.Show(
                message + " has changed.\nDo you want to save it?",
                "Confirmation", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
            switch (rslt)
            {
                case DialogResult.No:
                    return true;
                case DialogResult.Cancel:
                    return false;
                default:
                    miSave.PerformClick();
                    return !codeEditor.Modified;
            }
        }

        #endregion Form lifetime.

        #region File management.

        public string FileName
        {
            get { return fileName; }
            set
            {
                fileName = value;
                if (String.IsNullOrEmpty(fileName))
                {
                    this.Text = DefaultTitle + " - untitled";
                    printDoc.DocumentName = "Freya source file";
                }
                else
                {
                    this.Text = DefaultTitle + " - " + Path.GetFileName(fileName);
                    printDoc.DocumentName = fileName;
                }
            }
        }

        #endregion File management.

        #region File menu commands.

        private void miNew_Click(object sender, EventArgs e)
        {
            if (CanClose())
            {
                codeEditor.Reset();
                FileName = String.Empty;
                statusPanel.Text = String.Empty;
            }
        }

        private void miOpen_Click(object sender, EventArgs e)
        {
            if (CanClose() &&
                openFileDialog.ShowDialog() == DialogResult.OK)
            {
                codeEditor.Load(openFileDialog.FileName);
                FileName = openFileDialog.FileName;
                mruList.OpenFile(FileName);
                statusPanel.Text = String.Empty;
            }
        }

        private void mruList_FileOpen(object sender, MruOpenEventArgs e)
        {
            if (CanClose())
            {
                codeEditor.Load(e.FileName);
                FileName = e.FileName;
                // Move this file name to the top of the MRU list.
                mruList.OpenFile(FileName);
                statusPanel.Text = String.Empty;
            }
        }

        private void miSave_Click(object sender, EventArgs e)
        {
            if (String.IsNullOrEmpty(fileName))
                miSaveAs.PerformClick();
            else
            {
                codeEditor.Save(fileName);
                codeEditor.Modified = false;
            }
        }

        private void miSaveAs_Click(object sender, EventArgs e)
        {
            saveFileDialog.FileName = fileName;
            if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
            {
                codeEditor.Save(saveFileDialog.FileName);
                FileName = saveFileDialog.FileName;
                codeEditor.Modified = false;
                mruList.OpenFile(FileName);
            }
        }

        private void miPreview_Click(object sender, EventArgs e)
        {
            previewDialog.ShowDialog(this);
        }

        private void miPrint_Click(object sender, EventArgs e)
        {
            if (printDialog.ShowDialog() == DialogResult.OK)
                printDialog.Document.Print();
        }

        private void miFolder_Click(object sender, EventArgs e)
        {
            Process.Start("explorer.exe", Path.GetDirectoryName(FileName));
        }

        private void miProperties_Click(object sender, EventArgs e)
        {
            PropsDialog.Execute(FileName, codeEditor);
        }

        private void miExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        #endregion File menu commands.

        #region Edit menu commands.

        private void miEdit_DropDownOpening(object sender, EventArgs e)
        {
            miUndo.Text = codeEditor.CanUndo ? "&Undo " + codeEditor.UndoName : "&Undo";
            miRedo.Text = codeEditor.CanRedo ? "&Redo " + codeEditor.RedoName : "&Redo";
        }

        private void editorMenu_Opening(object sender, CancelEventArgs e)
        {
            ctUndo.Enabled = codeEditor.CanUndo;
            ctUndo.Text = codeEditor.CanUndo ? "&Undo " + codeEditor.UndoName : "&Undo";
            ctCut.Enabled = ctCopy.Enabled = ctDelete.Enabled = codeEditor.HasSelection;
            ctPaste.Enabled = codeEditor.CanPaste;
        }

        private void miUndo_Click(object sender, EventArgs e)
        {
            codeEditor.Undo();
        }

        private void miRedo_Click(object sender, EventArgs e)
        {
            codeEditor.Redo();
        }

        private void miCut_Click(object sender, EventArgs e)
        {
            codeEditor.Cut();
        }

        private void miCopy_Click(object sender, EventArgs e)
        {
            codeEditor.Copy();
        }

        private void miPaste_Click(object sender, EventArgs e)
        {
            codeEditor.Paste();
        }

        private void miDelete_Click(object sender, EventArgs e)
        {
            codeEditor.Delete();
        }

        private void miSelectAll_Click(object sender, EventArgs e)
        {
            codeEditor.SelectAll();
        }

        private void miGotoLine_Click(object sender, EventArgs e)
        {
            if (codeEditor.LineCount > 0)
                GotoForm.Execute(codeEditor);
        }

        private void miFind_Click(object sender, EventArgs e)
        {
            statusPanel.Text = String.Empty;
            FindDialog.Show(codeEditor,
                delegate(object sender1, FindCompleteEventArgs e1)
                {
                    statusPanel.Text = e1.Reason == FindResult.Found ? "Text found!" : "Not found!";
                });
        }

        private void miReplace_Click(object sender, EventArgs e)
        {
            statusPanel.Text = String.Empty;
            ReplaceDialog.Show(codeEditor,
                delegate(object sender1, FindCompleteEventArgs e1)
                {
                    switch (e1.Reason)
                    {
                        case FindResult.Found:
                            statusPanel.Text = "Text found.";
                            break;
                        case FindResult.NotFound:
                            statusPanel.Text = "Not found.";
                            break;
                        case FindResult.Replaced:
                            statusPanel.Text = "Replaced.";
                            break;
                        case FindResult.ReplacedAll:
                            statusPanel.Text = String.Format(
                                "Replaced {0} occurrence(s).", e1.Count);
                            break;
                    }
                });
        }

        private void miUpper_Click(object sender, EventArgs e)
        {
            codeEditor.ChangeCase(true);
        }

        private void miLower_Click(object sender, EventArgs e)
        {
            codeEditor.ChangeCase(false);
        }

        private void miComment_Click(object sender, EventArgs e)
        {
            codeEditor.CommentSelection(true);
        }

        private void miUncomment_Click(object sender, EventArgs e)
        {
            codeEditor.CommentSelection(false);
        }

        private void miSnippet_Click(object sender, EventArgs e)
        {
            codeEditor.ShowSnippets();
        }

        private void codeEditor_ModifiedChanged(object sender, EventArgs e)
        {
            modifiedLabel.Text = codeEditor.Modified ? "Modified" : String.Empty;
        }

        #endregion Edit menu commands.

        #region Project menu commands.

        private void miBuild_Click(object sender, EventArgs e)
        {

        }

        private void miRun_Click(object sender, EventArgs e)
        {

        }

        private void miOptions_Click(object sender, EventArgs e)
        {
            OptionsDialog.Execute(this);
        }

        #endregion Project menu commands.

        #region Help menu commands.

        private void miIntSight_Click(object sender, EventArgs e)
        {
            Process.Start("IExplore.exe", @"http://www.marteens.com/freya");
        }

        private void miMoebius_Click(object sender, EventArgs e)
        {
            Process.Start("IExplore.exe", @"http://www.moebiusootics.com");
        }

        private void miAbout_Click(object sender, EventArgs e)
        {
            AboutForm.Execute(this);
        }

        #endregion Help menu commands.
    }
}