using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Reflection;
using System.Reflection.Emit;
using System.Text;
using System.Windows.Forms;

namespace hammer
{
    public partial class OptionsDialog : Form
    {
        private static int lastTab = 0;

        public static bool Execute(Form owner)
        {
            using (OptionsDialog form = new OptionsDialog())
                if (form.ShowDialog(owner) == DialogResult.OK)
                {
                    form.SaveValues();
                    return true;
                }
                else
                    return false;
        }

        private OptionsDialog()
        {
            InitializeComponent();
            Opacity = (MainForm.GlobalOpacity - 1) / 100.0;
            foreach (FontFamily family in FontFamily.Families)
                if (family.IsStyleAvailable(FontStyle.Regular))
                    cbFamilies.Items.Add(family.Name);
            RestoreValues();
            tabControl.SelectedIndex = lastTab;
        }

        private void SaveValues()
        {
            fontWarning.Hide(cbFamilies);
            lastTab = tabControl.SelectedIndex;
            MainForm.WriteRegistry("Target kind",
                targetKind.SelectedIndex == 0 ? (int)PEFileKinds.ConsoleApplication :
                targetKind.SelectedIndex == 1 ? (int)PEFileKinds.WindowApplication :
                (int)PEFileKinds.Dll);
            MainForm.WriteRegistry("Entry type", entryType.Text);
            MainForm.WriteRegistry("Debug", debug.Checked);
            MainForm.WriteRegistry("Preconditions", preconditions.Checked);
            MainForm.WriteRegistry("Postconditions", postconditions.Checked);
            MainForm.WriteRegistry("Symbols", symbols.Text.Split(
                new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries));
            MainForm.WriteRegistry(MainForm.KeyVisualStyle, cbVisualStyle.SelectedIndex);
            MainForm.WriteRegistry(MainForm.KeyOpacity, opacity.Value);
            switch (cbVisualStyle.SelectedIndex)
            {
                case 0: MainForm.VStyle = MainForm.VisualStyle.Blue; break;
                case 1: MainForm.VStyle = MainForm.VisualStyle.Tan; break;
                case 2: MainForm.VStyle = MainForm.VisualStyle.Silver; break;
            }
            MainForm.GlobalOpacity = opacity.Value;
            MainForm.KeywordColor = colorPicker1.SelectedColor;
            MainForm.CommentColor = colorPicker2.SelectedColor;
            MainForm.StringColor = colorPicker3.SelectedColor;
            MainForm.WriteRegistry("Keywords color", MainForm.KeywordColor.ToArgb());
            MainForm.WriteRegistry("Comments color", MainForm.CommentColor.ToArgb());
            MainForm.WriteRegistry("Strings color", MainForm.StringColor.ToArgb());
            int tabSize;
            if (int.TryParse(edTabSize.Text, out tabSize))
                MainForm.TabSize = tabSize;
            MainForm.WriteRegistry(MainForm.KeyTabSize, MainForm.TabSize);
            int leftMargin;
            if (int.TryParse(edLeftMargin.Text, out leftMargin))
                MainForm.LeftMargin = leftMargin;
            MainForm.WriteRegistry(MainForm.KeyEditorMargin, MainForm.LeftMargin);
            MainForm.EditorFont = new Font(cbFamilies.Text, Single.Parse(cbFontSize.Text));
            MainForm.WriteRegistry("Editor font", MainForm.EditorFont.Name);
            MainForm.WriteRegistry("Editor font size", MainForm.EditorFont.Size.ToString());
        }

        private void RestoreValues()
        {
            switch ((PEFileKinds)
                MainForm.Read("Target kind", (int)PEFileKinds.ConsoleApplication))
            {
                case PEFileKinds.Dll:
                    targetKind.SelectedIndex = 2;
                    break;
                case PEFileKinds.WindowApplication:
                    targetKind.SelectedIndex = 1;
                    break;
                default:
                    targetKind.SelectedIndex = 0;
                    break;
            }
            entryType.Text = MainForm.Read("Entry type", String.Empty);
            debug.Checked = MainForm.Read("Debug", false);
            preconditions.Checked = MainForm.Read("Preconditions", false);
            postconditions.Checked = MainForm.Read("Postconditions", false);
            symbols.Text = String.Join(";", (string[])MainForm.ReadRegistry(
                "Symbols", new string[0]));
            switch (MainForm.VStyle)
            {
                case MainForm.VisualStyle.Silver:
                    cbVisualStyle.SelectedIndex = 2;
                    break;
                case MainForm.VisualStyle.Tan:
                    cbVisualStyle.SelectedIndex = 1;
                    break;
                default:
                    cbVisualStyle.SelectedIndex = 0;
                    break;
            }
            opacity.Value = MainForm.GlobalOpacity;
            edTabSize.Text = MainForm.TabSize.ToString();
            edLeftMargin.Text = MainForm.LeftMargin.ToString();
            colorPicker1.SelectedColor = MainForm.KeywordColor;
            colorPicker2.SelectedColor = MainForm.CommentColor;
            colorPicker3.SelectedColor = MainForm.StringColor;
            Font f = MainForm.EditorFont;
            cbFamilies.SelectedIndex = cbFamilies.FindString(f.Name);
            cbFontSize.SelectedIndex = cbFontSize.FindString(
                ((int)Math.Round(f.Size)).ToString());
        }

        private void opacity_ValueChanged(object sender, EventArgs e)
        {
            this.Opacity = (opacity.Value - 1) / 100.0;
        }

        private void edTabSize_KeyPress(object sender, KeyPressEventArgs e)
        {
            toolTip.Hide(edTabSize);
            toolTip.Hide(edLeftMargin);
            if (e.KeyChar >= 32)
                if (e.KeyChar < '0' || e.KeyChar > '9')
                {
                    ShowError((TextBox)sender, "Please, type a valid integer.");
                    e.Handled = true;
                }
        }

        private void OptionsDialog_FormClosed(object sender, FormClosedEventArgs e)
        {
            toolTip.Hide(edTabSize);
            toolTip.Hide(edLeftMargin);
        }

        private void ShowError(TextBox ctrl, string message)
        {
            toolTip.Show(message, ctrl, ctrl.Width, ctrl.Height, 3000);
        }

        private void OptionsDialog_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.DialogResult == DialogResult.OK)
            {
                int value;
                if (!int.TryParse(edTabSize.Text, out value))
                {
                    ShowError(edTabSize, "Tabulation size is not an integer value.");
                    edTabSize.Focus();
                    e.Cancel = true;
                }
                else if (value < 1)
                {
                    ShowError(edTabSize, "Tabulation size can't be lesser than 1.");
                    edTabSize.Focus();
                    e.Cancel = true;
                }
                if (!int.TryParse(edLeftMargin.Text, out value))
                {
                    ShowError(edTabSize, "Left margin is not an integer value.");
                    edLeftMargin.Focus();
                    e.Cancel = true;
                }
            }

        }

        private void edTabSize_Click(object sender, EventArgs e)
        {
            toolTip.Hide(edTabSize);
        }

        private void button1_Click(object sender, EventArgs e)
        {
            colorDialog.Color = colorPicker1.SelectedColor;
            if (colorDialog.ShowDialog(this) == DialogResult.OK)
                colorPicker1.SelectedColor = colorDialog.Color;
        }

        private void button2_Click(object sender, EventArgs e)
        {
            colorDialog.Color = colorPicker2.SelectedColor;
            if (colorDialog.ShowDialog(this) == DialogResult.OK)
                colorPicker2.SelectedColor = colorDialog.Color;
        }

        private void button3_Click(object sender, EventArgs e)
        {
            colorDialog.Color = colorPicker3.SelectedColor;
            if (colorDialog.ShowDialog(this) == DialogResult.OK)
                colorPicker3.SelectedColor = colorDialog.Color;
        }

        private bool warningAlreadyShown = false;

        private void cbFamilies_SelectionChangeCommitted(object sender, EventArgs e)
        {
            string fontname = cbFamilies.Text;
            if (String.Compare(fontname, "Courier", true) != 0 &&
                String.Compare(fontname, "Courier New", true) != 0 &&
                String.Compare(fontname, "CourierPS", true) != 0 &&
                String.Compare(fontname, "FixedSys", true) != 0 &&
                String.Compare(fontname, "Letter Gothic", true) != 0 &&
                String.Compare(fontname, "Lucida Console", true) != 0 &&
                String.Compare(fontname, "Lucida Sans Typewriter", true) != 0 &&
                String.Compare(fontname, "Terminal", true) != 0)
                if (!warningAlreadyShown)
                {
                    fontWarning.Show(
                        "The Code Editor only supports monospaced fonts.\n" +
                        "You have selected a font not recognized as such by this application.",
                        cbFamilies, 5000);
                    warningAlreadyShown = true;
                }
        }
    }
}