using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using IntSight.Controls;

namespace hammer
{
    public partial class ReplaceDialog : Form
    {
        private static ReplaceDialog instance = null;
        private IntSight.Controls.CodeEditor editor;
        private FindCompleteEventHandler saveFindComplete;

        public static void Show(
            IntSight.Controls.CodeEditor editor,
            FindCompleteEventHandler replaceComplete)
        {
            if (instance == null)
            {
                FindDialog.Shutdown();
                instance = new ReplaceDialog();
                instance.editor = editor;
                string text = editor.SelectedText;
                if (!string.IsNullOrEmpty(text))
                    instance.textCombo.Text = text;
                instance.saveFindComplete = replaceComplete;
                if (replaceComplete != null)
                    instance.FindComplete += replaceComplete;
                instance.Show(editor.FindForm());
            }
        }

        public static void SetOpacity()
        {
            if (instance != null)
                instance.Opacity = MainForm.GlobalOpacity;
        }

        public static void Shutdown()
        {
            if (instance != null)
                instance.Close();
        }

        public event FindCompleteEventHandler FindComplete;

        private ReplaceDialog()
        {
            InitializeComponent();
            Opacity = (MainForm.GlobalOpacity - 1) / 100.0;
            RestoreValues();
            FormClosed += ReplaceForm_FormClosed;
        }

        private void ReplaceForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            SaveValues();
            if (saveFindComplete != null)
                FindComplete -= saveFindComplete;
            instance = null;
        }

        private void SaveValues()
        {
            MainForm.WriteRegistry("Find text", textCombo.Text);
            MainForm.WriteRegistry("Replace text", replaceCombo.Text);
            MainForm.WriteRegistry("Match case", matchCase.Checked);
            MainForm.WriteRegistry("Whole words", wholeWords.Checked);
            MainForm.WriteRegistry("Find left position", Left);
            MainForm.WriteRegistry("Find top position", Top);
            string[] findStrings = new string[textCombo.Items.Count];
            textCombo.Items.CopyTo(findStrings, 0);
            MainForm.WriteRegistry("Find items", findStrings);
        }

        private void RestoreValues()
        {
            textCombo.Text = MainForm.Read("Find text", String.Empty);
            replaceCombo.Text = MainForm.Read("Replace text", String.Empty);
            matchCase.Checked = MainForm.Read("Match case", false);
            wholeWords.Checked = MainForm.Read("Whole words", false);
            int leftPos = MainForm.Read("Find left position", -1);
            int topPos = MainForm.Read("Find top position", -1);
            if (leftPos >= 0 && topPos >= 0)
            {
                this.StartPosition = FormStartPosition.Manual;
                this.Location = new Point(leftPos, topPos);
            }
            textCombo.Items.AddRange((string[])MainForm.ReadRegistry(
                "Find items", new string[0]));
        }

        protected virtual void OnFindComplete(FindCompleteEventArgs e)
        {
            FindCompleteEventHandler eh = FindComplete;
            if (eh != null)
                eh(this, e);
        }

        private void bnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void ReplaceForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((e.Modifiers & (Keys.Control | Keys.Shift | Keys.Alt)) == Keys.Control &&
                e.KeyCode == Keys.F)
            {
                FindDialog.Show(editor, FindComplete);
                e.Handled = true;
            }
        }

        private void bnFind_Click(object sender, EventArgs e)
        {
            string findString = textCombo.Text;
            textCombo.SelectAll();
            int pos = textCombo.Items.IndexOf(findString);
            if (pos != -1)
                textCombo.Items.RemoveAt(pos);
            if (textCombo.Items.Count == 8)
                textCombo.Items.RemoveAt(7);
            textCombo.Items.Insert(0, findString);
            bool result = editor.FindText(
                findString, editor.Current, wholeWords.Checked, matchCase.Checked);
            if (!result && editor.Current > CodeEditor.Position.Zero)
                result = editor.FindText(
                    findString, CodeEditor.Position.Zero, wholeWords.Checked, matchCase.Checked);
            OnFindComplete(result ?
                FindCompleteEventArgs.Found : FindCompleteEventArgs.NotFound);
        }

        private void bnReplace_Click(object sender, EventArgs e)
        {
            if (String.Compare(textCombo.Text, editor.SelectedText, !matchCase.Checked) == 0)
            {
                editor.SelectedText = replaceCombo.Text;
                OnFindComplete(FindCompleteEventArgs.Replaced);
                editor.FindText(textCombo.Text,
                    editor.Current, wholeWords.Checked, matchCase.Checked);
            }
            else
            {
                bool result = editor.FindText(textCombo.Text,
                    editor.Current, wholeWords.Checked, matchCase.Checked);
                OnFindComplete(result ?
                    FindCompleteEventArgs.Found : FindCompleteEventArgs.NotFound);
                if (!result)
                    FindDialog.Beep();
            }
        }

        private void bnReplaceAll_Click(object sender, EventArgs e)
        {
            CodeEditor.Position start = CodeEditor.Position.Zero;
            int count = 0;
            while (editor.FindText(
                textCombo.Text, start, wholeWords.Checked, matchCase.Checked))
            {
                start = editor.Current;
                editor.SelectedText = replaceCombo.Text;
                start.Column += replaceCombo.Text.Length - textCombo.Text.Length;
                count++;
            }
            if (count == 0)
            {
                OnFindComplete(FindCompleteEventArgs.NotFound);
                FindDialog.Beep();
            }
            else
                OnFindComplete(new FindCompleteEventArgs(count));
        }
    }
}